@@ -22,7 +22,7 @@ copyright = '2020-2021, Sean Whitton'
 author = 'Sean Whitton'
 
 # The full version, including alpha/beta/rc tags
-release = '0.8.0'
+release = '0.11.0'
 
 
 # -- General configuration ---------------------------------------------------
@@ -49,7 +49,7 @@ highlight_language = 'common-lisp'
 # The theme to use for HTML and HTML Help pages.  See the documentation for
 # a list of builtin themes.
 #
-html_theme = 'haiku'
+html_theme = 'sphinx_rtd_theme'
 
 # Add any paths that contain custom static files (such as style sheets) here,
 # relative to this directory. They are copied after the builtin static files,
