@@ -69,6 +69,8 @@ For an example of usage, see the :SUDO connection type."))
     :initform nil
     :documentation "This connection's connection attributes.")))
 
+(define-print-object-for-structlike connection)
+
 (defclass lisp-connection (connection) ())
 
 (defclass posix-connection (connection) ())
@@ -104,6 +106,23 @@ error condition just because EXIT is non-zero."))
   (let ((*connection* (slot-value connection 'parent)))
     (call-next-method)))
 
+(defgeneric connection-readfile-and-remove (connection path)
+  (:documentation "As READFILE and then delete the file.
+
+For some connection types, when latency is high, combining these two
+operations is noticeably faster than doing one after the other.  For every use
+of RUN we read and delete the file containing the command's stdout, so the
+time savings add up."))
+
+(defmethod connection-readfile-and-remove
+    :around ((connection connection) path)
+  (let ((*connection* (slot-value connection 'parent)))
+    (call-next-method)))
+
+(defmethod connection-readfile-and-remove ((connection connection) path)
+  (prog1 (connection-readfile connection path)
+    (connection-run connection (strcat "rm " (escape-sh-token path)) nil)))
+
 ;; only functional difference between WRITEFILE and UPLOAD is what args they
 ;; take: a string vs. a path.  for a given connection type, they may have same
 ;; or different implementations.
@@ -225,6 +244,12 @@ login(1)).  Tilde expansion works correctly."
         (failed-change "Failed to determine remote home directory.")
         (ensure-directory-pathname (stripln home)))))
 
+(defmethod connection-connattr
+    ((connection connection) (k (eql :XDG-CACHE-HOME)))
+  (ensure-directory-pathname
+   (stripln
+    (connection-run connection "echo ${XDG_CACHE_HOME:-$HOME/.cache}" nil))))
+
 
 ;;;; Functions to access the slots of the current connection
 
@@ -262,32 +287,57 @@ which will be cleaned up when BODY is finished."
     `(let ((,file (mktemp ,@(and directory-supplied-p
                                  `(:directory ,directory))
                           :connection ,connection)))
-       (unwind-protect-in-parent
-            (progn ,@body)
+       (unwind-protect (progn ,@body)
          (connection-run ,connection
                          (format nil "rm -f ~A" (escape-sh-token ,file))
                          nil)))))
 
+(defun mkstemp-cmd (&optional template
+                    &aux (template (or (unix-namestring template)
+                                       "'${TMPDIR:-/tmp}'/tmp.XXXXXX")))
+  ;; mktemp(1) is not POSIX; the only POSIX sh technique at the time of
+  ;; writing is to use m4(1)'s mkstemp macro.  However, m4 is sometimes not
+  ;; present, so fall back to mktemp(1).  Hopefully passing the template as
+  ;; the only command line option to mktemp(1) is portable.
+  ;;
+  ;; Although POSIX.1-2017 says that if m4(1) fails to create a temporary file
+  ;; it should exit nonzero, many m4(1) implementations just write to stderr
+  ;; and exit zero.  So we examine the stderr, and if there is any, exit
+  ;; nonzero ourselves.
+  ;;
+  ;; We apply the same stderr handling to mktemp(1), exiting if we see
+  ;; anything on stderr, as a simple way to ensure that non-fatal
+  ;; errors/warnings are not captured as the path to the temporary file.
+  ;;
+  ;; While GNU M4 mkstemp makes the temporary file at most readable and
+  ;; writeable by its owner, POSIX doesn't require this, so set a umask.
+  #?"umask 077
+exec 3>&1
+if err=\$(if command -v m4 >/dev/null; then
+              echo 'mkstemp(${template})' | m4 2>&1 1>&3
+          else
+              mktemp '${template}' 2>&1 1>&3
+          fi); then
+    case $err in
+        ?*) printf >&2 \"%s\\n\" \"$err\"; exit 1 ;;
+        *)  exit 0 ;;
+    esac
+else
+    case $err in
+        ?*) printf >&2 \"%s\\n\" \"$err\" ;;
+    esac
+    exit 1
+fi")
+
 (defun mktemp (&key (connection *connection*) directory)
   "Make a temporary file on the remote side, in DIRECTORY, defaulting to /tmp."
-  (let ((template (if directory
-                      (unix-namestring
-                       (merge-pathnames
-                        "tmp.XXXXXX" (ensure-directory-pathname directory)))
-                      "'${TMPDIR:-/tmp}'/tmp.XXXXXX")))
-    (multiple-value-bind (out exit)
-        ;; mktemp(1) is not POSIX; the only POSIX way is this M4 way,
-        ;; apparently, but even though m4(1) is POSIX it seems like it could
-        ;; often be absent, so have a fallback.  It would be better to avoid
-        ;; passing any arguments to mktemp(1) as these may differ on different
-        ;; platforms, but hopefully just a template is okay.
-        ;;
-        ;; While GNU M4 mkstemp makes the temporary file at most readable and
-        ;; writeable by its owner, POSIX doesn't require this, so set a umask.
-        (connection-run
-         connection
-         #?"umask 077; echo 'mkstemp(${template})' | m4 2>/dev/null || mktemp '${template}'"
-         nil)
+  (let ((cmd (format
+              nil "if tmpf=$(~A); then echo \"$tmpf\"; else exit 1; fi"
+              (mkstemp-cmd
+               (and directory
+                    (merge-pathnames
+                     "tmp.XXXXXX" (ensure-directory-pathname directory)))))))
+    (multiple-value-bind (out exit) (connection-run connection cmd nil)
       (let ((lines (lines out)))
         (if (and (zerop exit) lines)
             (car lines)
@@ -340,7 +390,8 @@ the working directory of the Lisp process using UIOP:WITH-CURRENT-DIRECTORY."
      (setq cmd (if (cdr cmd) (escape-sh-command cmd) (car cmd)))
      (loop while env
            collect (format nil "~A=~A"
-                           (symbol-name (pop env)) (escape-sh-token (pop env)))
+                           (string-upcase (symbol-name (pop env)))
+                           (escape-sh-token (pop env)))
              into accum
            finally
               (when accum
@@ -361,8 +412,9 @@ the working directory of the Lisp process using UIOP:WITH-CURRENT-DIRECTORY."
      ;; simplicity, particularly to avoid having to check whether the connattr
      ;; is set yet, because setting it requires working CONNECTION-RUN.
      (setq cmd (format nil "export HOME=~A; cd ~A; ~A"
-                       (escape-sh-token (unix-namestring
-                                         (get-connattr :remote-home)))
+                       (escape-sh-token (drop-trailing-slash
+                                         (unix-namestring
+                                          (get-connattr :remote-home))))
                        (escape-sh-token (unix-namestring (pwd)))
                        cmd))
      ,@forms))
@@ -397,17 +449,35 @@ Keyword arguments accepted:
 Returns command's stdout, stderr and exit code, unless :FOR-EXIT, in which
 case return only the exit code."
   (%process-run-args
-    (with-remote-temporary-file (stdout)
-      (setq cmd (format nil "( ~A ) >~A" cmd stdout))
-      (informat 4 "~&RUN ~A" cmd)
-      (multiple-value-bind (err exit)
-          (connection-run *connection* cmd input)
-        (let ((out (readfile stdout)))
-          (when inform (informat 1 "~&    % ~A~%~{    ~A~%~}" cmd (lines out)))
-          (if (or may-fail (= exit 0))
-              (if for-exit exit (values out err exit))
-              (error 'run-failed
-                     :cmd cmd :stdout out :stderr err :exit-code exit)))))))
+    (let (stdout
+          (wrapped
+            (format
+             nil "tmpf=$(~A) && printf \"%s\\n\" \"$tmpf\" && (~A) >\"$tmpf\""
+             (load-time-value (mkstemp-cmd) t) cmd)))
+      (handler-bind
+          ((serious-condition
+             (lambda (c)
+               (declare (ignore c))
+               (when stdout
+                 (connection-run
+                  *connection*
+                  (format nil "rm -f ~A" (escape-sh-token stdout))
+                  nil)))))
+        (informat 4 "~&RUN ~A"
+                  (if (> *consfigurator-debug-level* 4) wrapped cmd))
+        (multiple-value-bind (err exit)
+            (connection-run *connection* wrapped input)
+          (setq err (lines err) stdout (car err) err (unlines (cdr err)))
+          (let ((out (connection-readfile-and-remove *connection* stdout)))
+            (when inform
+              (informat 1 "~&    % ~A~%~{    ~A~%~}"
+                        (if (> *consfigurator-debug-level* 4) wrapped cmd)
+                        (lines out)))
+            (if (or may-fail (= exit 0))
+                (if for-exit exit (values out err exit))
+                (error 'run-failed
+                       :cmd (if (> *consfigurator-debug-level* 4) wrapped cmd)
+                       :stdout out :stderr err :exit-code exit))))))))
 
 (defun mrun (&rest args)
   "Like RUN but don't separate stdout and stderr (\"m\" for \"merged\"; note
@@ -443,10 +513,22 @@ expected."
 (defun test (&rest args)
   (zerop (apply #'mrun :for-exit "test" args)))
 
+(defun mountpointp (path)
+  "Is PATH a mount point?
+
+Uses mountpoint(1) from util-linux, so add a property requiring OS:LINUX or a
+subclass to the :HOSTATTRS subroutine of properties calling this."
+  (zerop (mrun :for-exit "mountpoint" "-q" path)))
+
 (defun delete-remote-trees (&rest paths)
   "Recursively delete each of PATHS."
   (mrun "rm" "-rf" paths))
 
+(defun empty-remote-directory (directory)
+  "Recursively delete the contents of DIRECTORY, but not DIRECTORY itself."
+  (alet (escape-sh-token (drop-trailing-slash (unix-namestring directory)))
+    (mrun (format nil "rm -rf -- ~A/* ~A/.[!.]* ~A/..?*" it it it))))
+
 (defun remote-exists-p (&rest paths)
   "Does each of PATHS exists?
 PATH may be any kind of file, including directories."
@@ -454,8 +536,14 @@ PATH may be any kind of file, including directories."
                       nconc (list "-e" (car path))
                       when (cdr path) collect "-a")))
 
-(defun remote-file-mode-and-size (path)
-  "Get the numeric mode and size in bytes of PATH, or NIL if it does not exist."
+(defun remote-file-stats (path)
+  "Get the numeric mode, size in bytes and mtime of PATH, or NIL if it does not
+exist.
+
+The mtime is only accurate to the nearest UTC day, rounding down, if the file
+was modified in the past six months or its mtime is in the future, and only
+accurate to the nearest minute, rounding down, otherwise (see the
+specification of POSIX ls(1))."
   (flet ((sum (chars order)
            (+ (if (char= (elt chars 0) #\r) (* order 4) 0)
               (if (char= (elt chars 1) #\w) (* order 2) 0)
@@ -467,12 +555,51 @@ PATH may be any kind of file, including directories."
                 (#\x order)
                 (#\- 0)))))
     (and (remote-exists-p path)
-         (let* ((ls (split-string (run "ls" "-ld" path)))
+         ;; This is a safe parse of ls(1) given its POSIX specification.
+         (let* ((ls (words
+                     (run :env '(:LC_ALL "C" :TZ "UTC") "ls" "-ld" path)))
                 (lscar (car ls)))
            (values (+ (sum (subseq lscar 1 4) #o100)
                       (sum (subseq lscar 4 7) #o10)
                       (sum (subseq lscar 7 10) 1))
-                   (parse-integer (nth 4 ls)))))))
+                   (parse-integer (nth 4 ls))
+                   (let ((date (parse-integer (nth 6 ls)))
+                         (month (cdr
+                                 (assoc
+                                  (nth 5 ls)
+                                  '(("Jan" . 1) ("Feb" . 2) ("Mar" . 3)
+                                    ("Apr" . 4) ("May" . 5) ("Jun" . 6)
+                                    ("Jul" . 7) ("Aug" . 8) ("Sep" . 9)
+                                    ("Oct" . 10) ("Nov" . 11) ("Dec" . 12))
+                                  :test #'string=))))
+                     (if (find #\: (nth 7 ls))
+                         (destructuring-bind (hour minute)
+                             (split-string (nth 7 ls) :separator ":")
+                           (encode-universal-time
+                            0 (parse-integer minute) (parse-integer hour)
+                            date month (nth-value 5 (get-decoded-time))
+                            0))
+                         (encode-universal-time
+                          0 0 0 date month (parse-integer (nth 7 ls)) 0))))))))
+
+(defun remote-last-reboot ()
+  "Get the time of the last reboot, rounded down to the nearest minute."
+  ;; The '-b' option to who(1) is specified in POSIX, though not the output
+  ;; format; this parse is based on GNU coreutils who(1).
+  (multiple-value-bind (match groups)
+      (re:scan-to-strings
+       "([0-9]{4})-([0-9]{2})-([0-9]{2}) ([0-9]{2}):([0-9]{2})"
+       (car (runlines :env '(:TZ "UTC") "who" "-b")))
+    (if match
+        (let ((groups (map 'vector #'parse-integer groups)))
+          (encode-universal-time 0 (elt groups 4) (elt groups 3)
+                                 (elt groups 2) (elt groups 1) (elt groups 0)
+                                 0))
+        (failed-change "Could not determine time of remote's last reboot."))))
+
+(defun remote-consfigurator-cache-pathname (path)
+  (merge-pathnames
+   path (merge-pathnames "consfigurator/" (get-connattr :XDG-CACHE-HOME))))
 
 (defun readfile (path)
   (connection-readfile
