@@ -21,12 +21,10 @@
 ;; currently we only check whether we're root, but, for example, on Linux, we
 ;; might have a CAP_* which lets us chroot as non-root
 (defun can-chroot ()
-  (zerop (foreign-funcall "geteuid" :int)))
+  (zerop (nix:geteuid)))
 
 (defmethod establish-connection ((type (eql :chroot)) remaining &key into)
-  (establish-connection (if (and (lisp-connection-p)
-                                 (can-chroot)
-                                 (can-probably-fork))
+  (establish-connection (if (and (lisp-connection-p) (can-chroot))
                             :chroot.fork
                             :chroot.shell)
                         remaining
@@ -48,7 +46,7 @@ should be the mount point, without the chroot's root prefixed.")
                                  (slot-value connection 'into))))
       ;; We only mount when the target is not already a mount point, so we
       ;; don't shadow anything that the user has already set up.
-      (when (plusp (mrun :for-exit "mountpoint" "-q" dest))
+      (unless (mountpointp dest)
         (setq mount-args (copy-list mount-args))
         (setf (lastcar mount-args) dest)
         (apply #'mrun "mount" mount-args)
@@ -56,30 +54,25 @@ should be the mount point, without the chroot's root prefixed.")
 
 (defmethod connection-teardown :before ((connection chroot-connection))
   (dolist (mount (chroot-mounts connection))
-    (mrun "umount" mount)))
-
-(defparameter *standard-chroot-mounts* '(
-("-t" "proc"     "-o" "nosuid,noexec,nodev"                "proc"   "/proc")
-("-t" "sysfs"    "-o" "nosuid,noexec,nodev,ro"             "sys"    "/sys")
-("-t" "devtmpfs" "-o" "mode=0755,nosuid"                   "udev"   "/dev")
-("-t" "devpts"   "-o" "mode=0620,gid=5,nosuid,noexec"      "devpts" "/dev/pts")
-("-t" "tmpfs"    "-o" "mode=1777,nosuid,nodev"             "shm"    "/dev/shm")
-("-t" "tmpfs"    "-o" "mode=1777,strictatime,nodev,nosuid" "tmp"    "/tmp")
-("--bind"                                                  "/run"   "/run")))
+    ;; There shouldn't be any processes left running in the chroot after we've
+    ;; finished deploying it, but it's quite easy to end up with things like
+    ;; gpg-agent holding on to /dev/null, for example, so for simplicity, do a
+    ;; lazy unmount.
+    (mrun "umount" "-l" mount)))
 
 (defmethod initialize-instance :after ((connection chroot-connection) &key)
   (when (string= "Linux" (stripln (run "uname")))
     (with-slots (into) connection
-      ;; Ensure the chroot itself is a mountpoint so that findmnt(1) works
+      ;; Ensure the chroot itself is a mountpoint so that findmnt(8) works
       ;; correctly within the chroot.
-      (unless (zerop (mrun :for-exit "mountpoint" "-q" into))
-        (chroot-mount connection "--bind" into "/"))
+      (unless (mountpointp into) (chroot-mount connection "--bind" into "/"))
       ;; Now set up the usual bind mounts.  Help here from arch-chroot(8).
-      (dolist (mount *standard-chroot-mounts*)
+      (mount:assert-devtmpfs-udev-/dev)
+      (dolist (mount mount:*standard-linux-vfs*)
         (apply #'chroot-mount connection mount))
+      (chroot-mount connection "--bind" "/run" "/run")
       (when (remote-exists-p "/sys/firmware/efi/efivars")
-        (chroot-mount connection "-t" "efivarfs" "-o" "nosuid,noexec,nodev"
-                      "efivarfs" "/sys/firmware/efi/efivars")))))
+        (apply #'chroot-mount connection mount:*linux-efivars-vfs*)))))
 
 (defmethod propagate-connattr
     ((type (eql :opened-volumes)) connattr (connection chroot-connection))
@@ -88,10 +81,9 @@ should be the mount point, without the chroot's root prefixed.")
           when (and (subtypep (type-of volume) 'disk:filesystem)
                     (slot-boundp volume 'disk:mount-point)
                     (subpathp (disk:mount-point volume) into))
-            collect (let ((copy (disk:copy-volume-and-contents volume)))
-                      (setf (disk:mount-point copy)
-                            (in-chroot-pathname (disk:mount-point copy) into))
-                      copy)
+            collect (aprog1 (disk:copy-volume-and-contents volume)
+                      (setf (disk:mount-point it)
+                            (in-chroot-pathname (disk:mount-point it) into)))
           else collect volume)))
 
 (defmethod propagate-connattr
@@ -105,41 +97,33 @@ should be the mount point, without the chroot's root prefixed.")
 
 ;;;; :CHROOT.FORK
 
-(defun chroot (path)
-  #+sbcl      (sb-posix:chroot path)
-  #-(or sbcl) (foreign-funcall "chroot" :string path :int))
-
 (defclass chroot.fork-connection
     (rehome-connection chroot-connection fork-connection) ())
 
 (defmethod establish-connection ((type (eql :chroot.fork)) remaining &key into)
-  (unless (and (lisp-connection-p) (zerop (foreign-funcall "geteuid" :int)))
+  (unless (and (lisp-connection-p) (zerop (nix:geteuid)))
     (error "~&Forking into a chroot requires a Lisp image running as root"))
   (informat 1 "~&Forking into chroot at ~A" into)
-  (let* ((into* (ensure-directory-pathname into))
-         (connection (make-instance 'shell-chroot-connection :into into*)))
-    ;; This has the side effect of populating the CONSFIGURATOR::ID and
-    ;; :REMOTE-HOME connattrs correctly, so that they don't get bogus values
-    ;; when this connection object is used in UPLOAD-ALL-PREREQUISITE-DATA.
-    (multiple-value-bind (datadir-inside exit)
-        (connection-run
-         connection
-         (format nil "echo ${XDG_CACHE_HOME:-~A/.cache}/consfigurator/data/"
-                 (connection-connattr connection :remote-home))
-         nil)
-      (unless (zerop exit)
-        (error "Failed to determine datadir inside chroot."))
-      (setq connection (change-class connection 'chroot.fork-connection))
-      (setf (slot-value connection 'datadir)
-            (ensure-pathname
-             (stripln (subseq datadir-inside 1))
-             :defaults into* :ensure-absolute t :ensure-directory t))
-      (unwind-protect-in-parent (continue-connection connection remaining)
-        (connection-teardown connection)))))
+  (let* ((into (ensure-pathname into :want-absolute t :ensure-directory t))
+         (connection (make-instance 'shell-chroot-connection :into into)))
+    ;; Populate the CONSFIGURATOR::ID and :REMOTE-HOME connattrs correctly to
+    ;; ensure they don't get bogus values when this connection object is used
+    ;; in UPLOAD-ALL-PREREQUISITE-DATA.
+    (connection-connattr connection :remote-home)
+    ;; Obtain & cache XDG_CACHE_HOME inside the chroot, and compute DATADIR.
+    (let ((xdg-cache-home (connection-connattr connection :XDG-CACHE-HOME)))
+      (setf connection (change-class connection 'chroot.fork-connection)
+            (slot-value connection 'datadir)
+            (merge-pathnames
+             "consfigurator/data/" (chroot-pathname xdg-cache-home into))))
+    (unwind-protect (continue-connection connection remaining)
+      (connection-teardown connection))))
 
 (defmethod post-fork ((connection chroot.fork-connection))
-  (unless (zerop (chroot (slot-value connection 'into)))
-    (error "chroot(2) failed!"))
+  (with-slots (into) connection
+    (unless (zerop
+             (foreign-funcall "chroot" :string (unix-namestring into) :int))
+      (error "chroot(2) failed!")))
   (let ((home (connection-connattr connection :remote-home)))
     (setf (getenv "HOME") (unix-namestring home))
     ;; chdir, else our current working directory is a pointer to something
