@@ -20,7 +20,21 @@
 
 ;;;; Deployments
 
-(defun %consfigure (connections host)
+(defparameter *at-end-functions* nil)
+
+(defun at-end (function)
+  "Request that FUNCTION be called at the end of the current (sub)deployment.
+Called by property :APPLY and :UNAPPLY subroutines.  FUNCTION will be passed a
+single argument representing whether or not the deployment made a change.
+
+Properties which call this are responsible for ensuring that the I/O performed
+by FUNCTION is compatible with the connection type.  This amounts to the
+following requirement: if FUNCTION performs I/O beyond what :POSIX property
+:APPLY subroutines are permitted to perform, the property calling AT-END to
+register FUNCTION must be declared to be a :LISP property."
+  (push (ensure-function function) *at-end-functions*))
+
+(defun %consfigure (connections host &key (collect-at-end t))
   "Consfigurator's primary loop, recursively binding *CONNECTION* and *HOST*.
 
 Assumes arguments to connections in CONNECTIONS have been both normalised and
@@ -29,7 +43,12 @@ preprocessed."
       ((apply-*host*-propspec ()
          (let ((propapp (eval-propspec (host-propspec *host*))))
            (assert-connection-supports (propapptype propapp))
-           (propappapply propapp)))
+           (if collect-at-end
+               (let (*at-end-functions*)
+                 (let ((result (propappapply propapp)))
+                   (dolist (function *at-end-functions* result)
+                     (funcall function result))))
+               (propappapply propapp))))
        (connect (connections)
          (destructuring-bind ((type . args) . remaining) connections
            ;; implementations of ESTABLISH-CONNECTION which call
@@ -38,7 +57,7 @@ preprocessed."
            (multiple-value-bind (*connection* return)
                (apply #'establish-connection type remaining args)
              (if *connection*
-                 (unwind-protect-in-parent
+                 (unwind-protect
                      (if remaining (connect remaining) (apply-*host*-propspec))
                    (connection-teardown *connection*))
                  return)))))
@@ -53,7 +72,7 @@ preprocessed."
         (t
          (connect '((:local))))))))
 
-(defun consfigure (propspec-expression)
+(defun consfigure (propspec-expression &key collect-at-end)
   "Immediately preprocess and apply PROPSPEC-EXPRESSION in the context of the
 current target host and connection.  This function is provided for use by
 specialised property combinators.  It should not be used in property
@@ -69,7 +88,8 @@ will not be discarded."
    nil (make-host
         :hostattrs (hostattrs *host*)
         :propspec (with-*host*-*consfig*
-                    (make-propspec :propspec propspec-expression)))))
+                    (make-propspec :propspec propspec-expression)))
+   :collect-at-end collect-at-end))
 
 (defun deploy* (connections host &optional additional-properties)
   "Execute the deployment which is defined by the pair (CONNECTIONS . HOST),
@@ -80,8 +100,9 @@ This is the entry point to Consfigurator's primary loop.  Typically users use
 DEPLOY, DEPLOY-THESE, and the function definitions established by DEFDEPLOY,
 DEFDEPLOY-THESE, etc., rather than calling this function directly.  However,
 code which programmatically constructs deployments will need to call this."
-  (%consfigure (preprocess-connections connections)
-               (union-propspec-into-host host additional-properties)))
+  (with-deployment-report
+      (%consfigure (preprocess-connections connections)
+                   (union-propspec-into-host host additional-properties))))
 
 (defun deploy-these* (connections host properties)
   "Like DEPLOY*, but replace the properties of HOST with PROPERTIES.
@@ -91,8 +112,9 @@ properties, plus any set by PROPERTIES.  Static informational attributes set
 by PROPERTIES can override the host's usual static informational attributes,
 in the same way that later entries in the list of properties specified in
 DEFHOST forms can override earlier entries (see DEFHOST's docstring)."
-  (%consfigure (preprocess-connections connections)
-               (replace-propspec-into-host host properties)))
+  (with-deployment-report
+      (%consfigure (preprocess-connections connections)
+                   (replace-propspec-into-host host properties))))
 
 (defun continue-deploy* (connection remaining-connections)
   "Complete the work of an enclosing call to DEPLOY* or DEPLOY-THESE*.
@@ -219,7 +241,8 @@ for testing properties at the REPL.  See also EVALS."
                         (run-program '("id") :output :string)))
             (,hostname (hostname-f))
             (,host (or (symbol-value (find-symbol (string-upcase ,hostname)))
-                       (make-host :hostattrs `(:hostname (,,hostname))))))
+                       (make-host :hostattrs `(:hostname (,,hostname))
+                                  :propspec (make-propspec :systems nil)))))
        (deploy-these*
         `((:sudo :as ,(format nil "~A@~A" ,username ,hostname)))
         ,host
@@ -270,7 +293,8 @@ PROPERTIES, like DEPLOY-THESE."
   (:desc (declare (ignore connections host properties)) "Subdeployment")
   (:preprocess
    (list (preprocess-connections connections)
-         (preprocess-host (replace-propspec-into-host host properties))
+         (preprocess-host
+          (replace-propspec-into-host (ensure-host host) properties))
          nil))
   (:hostattrs
    (declare (ignore connections properties))
