@@ -1,13 +1,14 @@
 (in-package :cl-user)
 
 (defpackage :consfigurator
-  (:use #:cl #:alexandria)
+  (:use #:cl #:anaphora #:alexandria #:cffi)
   (:local-nicknames (#:re #:cl-ppcre))
   (:shadowing-import-from #:uiop
                           #:strcat
                           #:string-prefix-p
                           #:string-suffix-p
                           #:split-string
+                          #:first-char
                           #:last-char
                           #:run-program
                           #:read-file-string
@@ -18,18 +19,26 @@
                           #:unix-namestring
                           #:pathname-directory-pathname
                           #:pathname-parent-directory-pathname
+                          #:resolve-symlinks
                           #:with-temporary-file
                           #:ensure-directory-pathname
                           #:ensure-pathname
                           #:enough-pathname
+                          #:pathname-equal
                           #:subpathp
+                          #:relative-pathname-p
                           #:getenv
                           #:subdirectories
                           #:directory-files
                           #:file-exists-p
 			  #:directory-exists-p
+                          #:rename-file-overwriting-target
                           #:with-current-directory
+                          #:delete-empty-directory
 			  #:delete-directory-tree
+                          #:with-safe-io-syntax
+                          #:slurp-stream-form
+                          #:safe-read-file-form
                           #:safe-read-from-string
                           #:compile-file*
                           #:compile-file-pathname*)
@@ -38,6 +47,7 @@
            #:string-prefix-p
            #:string-suffix-p
            #:split-string
+           #:first-char
            #:last-char
            #:run-program
            #:read-file-string
@@ -48,18 +58,26 @@
            #:unix-namestring
            #:pathname-directory-pathname
            #:pathname-parent-directory-pathname
+           #:resolve-symlinks
            #:with-temporary-file
            #:ensure-directory-pathname
            #:ensure-pathname
            #:enough-pathname
+           #:pathname-equal
            #:subpathp
+           #:relative-pathname-p
            #:getenv
            #:subdirectories
            #:directory-files
            #:file-exists-p
 	   #:directory-exists-p
+           #:rename-file-overwriting-target
            #:with-current-directory
+           #:delete-empty-directory
 	   #:delete-directory-tree
+           #:with-safe-io-syntax
+           #:slurp-stream-form
+           #:safe-read-file-form
            #:safe-read-from-string
            #:compile-file*
            #:compile-file-pathname*
@@ -76,6 +94,8 @@
            #:plist-to-cmd-args
 	   #:with-local-temporary-directory
            #:pathname-file
+           #:directory-contents
+           #:ensure-trailing-slash
            #:drop-trailing-slash
            #:quote-nonselfeval
            #:define-print-object-for-structlike
@@ -84,6 +104,9 @@
            #:escape-sh-token
            #:escape-sh-command
            #:defpackage-consfig
+           #:lambda-ignoring-args
+           #:parse-cidr
+           #:system
 
            #:*consfigurator-debug-level*
            #:with-indented-inform
@@ -98,8 +121,9 @@
            #:string->filename
            #:filename->string
 
-           #:unwind-protect-in-parent
-           #:cancel-unwind-protect-in-parent-cleanup
+           #:with-backtrace-and-exit-code
+           #:return-exit
+           #:posix-login-environment
 
            ;; connection.lisp
            #:establish-connection
@@ -112,6 +136,7 @@
            #:lisp-connection-p
            #:connection-run
            #:connection-readfile
+           #:connection-readfile-and-remove
            #:connection-writefile
            #:connection-teardown
            #:connection-connattr
@@ -120,13 +145,19 @@
            #:run
            #:mrun
            #:with-remote-temporary-file
+           #:mkstemp-cmd
+           #:mktemp
            #:with-remote-current-directory
            #:run-failed
            #:runlines
            #:test
            #:remote-exists-p
-           #:remote-file-mode-and-size
+           #:remote-file-stats
+           #:remote-last-reboot
+           #:remote-consfigurator-cache-pathname
+           #:mountpointp
            #:delete-remote-trees
+           #:empty-remote-directory
            #:readfile
            #:writefile
            #:get-connattr
@@ -136,7 +167,9 @@
            #:propattrs
            #:propunapply
            #:collapse-types
+           #:collapse-propapp-types
            #:propapptype
+           #:propappargs
            #:propappdesc
            #:propappattrs
            #:propappcheck
@@ -152,6 +185,7 @@
            #:get-parent-hostattrs
            #:get-parent-hostattrs-car
            #:push-hostattrs
+           #:pushnew-hostattr
            #:pushnew-hostattrs
            #:get-hostname
            #:get-short-hostname
@@ -162,9 +196,15 @@
            #:maybe-writefile-string
            #:call-with-os
            #:with-change-if-changes-file
+           #:with-change-if-changes-files
            #:with-change-if-changes-file-content
            #:with-change-if-changes-file-content-or-mode
 
+           #:ptype
+           #:plambda
+           #:papply
+           #:punapply
+
            ;; propspec.lisp
            #:in-consfig
            #:propspec-systems
@@ -183,8 +223,10 @@
            #:unapply
            #:desc
            #:on-change
+           #:on-apply-change
 	   #:as
            #:with-flagfile
+           #:with-unapply
 
            ;; host.lisp
            #:host
@@ -202,6 +244,7 @@
            #:with-replace-hostattrs
 
            ;; deployment.lisp
+           #:at-end
            #:consfigure
            #:defdeploy
            #:defdeploy-these
@@ -241,6 +284,7 @@
            #:get-remote-cached-prerequisite-data
            #:get-local-cached-prerequisite-data
 	   #:get-highest-local-cached-prerequisite-data
+           #:get-remote-data-cache-dir
 
            #:try-register-data-source
            #:register-data-source
@@ -255,6 +299,11 @@
            #:passphrase
            #:make-passphrase
            #:get-data-protected-string
+
+           ;; image.lisp
+           #:eval-in-grandchild
+           #:dump-consfigurator-in-grandchild
+           #:image-dumped
            #:asdf-requirements-for-host-and-features
            #:request-asdf-requirements
            #:continue-deploy*-program))
@@ -268,14 +317,19 @@
   (:local-nicknames (#:re #:cl-ppcre))
   (:export #:map-file-lines
            #:has-content
+           #:exists-with-content
            #:contains-lines
            #:lacks-lines
+           #:lacks-lines-matching
            #:has-mode
+           #:has-ownership
            #:does-not-exist
+           #:directory-does-not-exist
            #:data-uploaded
            #:host-data-uploaded
            #:secret-uploaded
            #:host-secret-uploaded
+           #:data-cache-purged
            #:contains-conf-equals
            #:contains-conf-space
            #:contains-conf-tab
@@ -305,6 +359,7 @@
            #:debian-stable
            #:debian-testing
            #:debian-unstable
+           #:debian-experimental
            #:debian-suite
            #:debian-architecture
            #:typecase
@@ -319,6 +374,11 @@
   (:export #:contained
            #:when-contained))
 
+(defpackage :consfigurator.property.periodic
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:file  #:consfigurator.property.file))
+  (:export #:at-most))
+
 (defpackage :consfigurator.property.mount
   (:use #:cl #:alexandria #:consfigurator)
   (:local-nicknames (#:os    #:consfigurator.property.os)
@@ -326,7 +386,11 @@
                     (#:file  #:consfigurator.property.file))
   (:export #:mounted
            #:unmounted-below
-           #:unmounted-below-and-removed))
+           #:unmounted-below-and-removed
+           #:all-mounts
+           #:*standard-linux-vfs*
+           #:*linux-efivars-vfs*
+           #:assert-devtmpfs-udev-/dev))
 
 (defpackage :consfigurator.property.service
   (:use #:cl #:alexandria #:consfigurator)
@@ -334,6 +398,8 @@
                     (#:file  #:consfigurator.property.file))
   (:export #:no-services
            #:running
+           #:restarted
+           #:reloaded
            #:without-starting-services))
 
 (defpackage :consfigurator.property.apt
@@ -344,22 +410,31 @@
                     (#:os         #:consfigurator.property.os)
                     (#:service    #:consfigurator.property.service))
   (:export #:installed
+           #:installed-minimally
            #:removed
            #:reconfigured
            #:service-installed-running
            #:all-configured
            #:updated
+           #:upgraded
+           #:autoremoved
            #:periodic-updates
            #:unattended-upgrades
            #:mirror
-           #:uses-parent-mirror
+           #:uses-parent-mirrors
            #:proxy
            #:uses-parent-proxy
            #:uses-local-cacher
+           #:get-mirrors
            #:standard-sources.list
+           #:additional-sources
            #:cache-cleaned
+           #:trusts-key
            #:all-installed-p
-           #:none-installed-p))
+           #:none-installed-p
+           #:suites-available-pinned
+           #:pinned
+           #:no-pdiffs))
 
 (defpackage :consfigurator.connection.sbcl
   (:use #:cl #:alexandria #:consfigurator)
@@ -368,11 +443,15 @@
 
 (defpackage :consfigurator.property.user
   (:use #:cl #:consfigurator)
-  (:local-nicknames (#:file  #:consfigurator.property.file))
+  (:local-nicknames (#:file  #:consfigurator.property.file)
+                    (#:os    #:consfigurator.property.os))
   (:export #:has-account
+           #:has-account-with-uid
            #:has-groups
+           #:has-desktop-groups
 	   #:has-login-shell
            #:has-enabled-password
+           #:has-locked-password
 	   #:passwd-entry))
 
 (defpackage :consfigurator.property.chroot
@@ -381,6 +460,7 @@
                     (#:apt       #:consfigurator.property.apt)
                     (#:os        #:consfigurator.property.os)
                     (#:container #:consfigurator.property.container)
+                    (#:mount     #:consfigurator.property.mount)
                     (#:file      #:consfigurator.property.file))
   (:shadow #:deploys #:deploys. #:deploys-these #:deploys-these.)
   (:export #:deploys
@@ -392,21 +472,11 @@
            #:os-bootstrapped
            #:os-bootstrapped.))
 
-(defpackage :consfigurator.property.live-build
-  (:use #:cl #:alexandria #:consfigurator)
-  (:local-nicknames (#:apt       #:consfigurator.property.apt)
-                    (#:os        #:consfigurator.property.os)
-                    (#:file      #:consfigurator.property.file)
-                    (#:mount     #:consfigurator.property.mount)
-                    (#:chroot    #:consfigurator.property.chroot))
-  (:export #:installed
-           #:image-built
-           #:image-built.))
-
 (defpackage :consfigurator.property.disk
-  (:use #:cl #:alexandria #:consfigurator)
+  (:use #:cl #:anaphora #:alexandria #:consfigurator)
   (:local-nicknames (#:re      #:cl-ppcre)
                     (#:chroot  #:consfigurator.property.chroot)
+                    (#:cmd     #:consfigurator.property.cmd)
                     (#:file    #:consfigurator.property.file)
                     (#:os      #:consfigurator.property.os)
                     (#:apt     #:consfigurator.property.apt))
@@ -458,6 +528,8 @@
            #:has-volumes
            #:caches-cleaned
            #:raw-image-built-for
+           #:debian-live-iso-built
+           #:debian-live-iso-built.
            #:host-volumes-created
            #:host-logical-volumes-exist
 
@@ -465,7 +537,11 @@
            #:volumes))
 
 (defpackage :consfigurator.property.fstab
-  (:use #:cl #:alexandria #:consfigurator #:consfigurator.property.disk)
+  (:use #:cl
+        #:anaphora
+        #:alexandria
+        #:consfigurator
+        #:consfigurator.property.disk)
   (:local-nicknames (#:os    #:consfigurator.property.os)
                     (#:file  #:consfigurator.property.file))
   (:export #:volume->entry
@@ -474,7 +550,11 @@
            #:entries-for-opened-volumes))
 
 (defpackage :consfigurator.property.crypttab
-  (:use #:cl #:alexandria #:consfigurator #:consfigurator.property.disk)
+  (:use #:cl
+        #:anaphora
+        #:alexandria
+        #:consfigurator
+        #:consfigurator.property.disk)
   (:local-nicknames (#:re    #:cl-ppcre)
                     (#:os    #:consfigurator.property.os)
                     (#:file  #:consfigurator.property.file))
@@ -483,19 +563,28 @@
 
 (defpackage :consfigurator.property.gnupg
   (:use #:cl #:consfigurator)
-  (:export #:public-key-imported))
+  (:local-nicknames (#:re        #:cl-ppcre))
+  (:export #:public-key-imported
+           #:secret-key-imported))
 
 (defpackage :consfigurator.property.git
   (:use #:cl #:consfigurator)
-  (:local-nicknames (#:file      #:consfigurator.property.file))
-  (:export #:snapshot-extracted))
+  (:local-nicknames (#:os        #:consfigurator.property.os)
+                    (#:file      #:consfigurator.property.file)
+                    (#:apt       #:consfigurator.property.apt))
+  (:export #:installed
+           #:snapshot-extracted
+           #:cloned
+           #:pulled
+           #:repo-configured))
 
 (defpackage :consfigurator.property.sshd
   (:use #:cl #:consfigurator)
   (:local-nicknames (#:re        #:cl-ppcre)
                     (#:os        #:consfigurator.property.os)
                     (#:file      #:consfigurator.property.file)
-                    (#:apt       #:consfigurator.property.apt))
+                    (#:apt       #:consfigurator.property.apt)
+                    (#:service   #:consfigurator.property.service))
   (:export #:installed
            #:configured
            #:no-passwords
@@ -508,6 +597,7 @@
   (:local-nicknames (#:file      #:consfigurator.property.file)
                     (#:sshd      #:consfigurator.property.sshd))
   (:export #:authorized-keys
+           #:has-user-key
            #:known-host
            #:globally-known-host
            #:parent-is-globally-known-host))
@@ -522,17 +612,27 @@
   (:export #:available
            #:selected-for))
 
+(defpackage :consfigurator.property.reboot
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:container    #:consfigurator.property.container))
+  (:export #:rebooted-at-end))
+
 (defpackage :consfigurator.property.installer
   (:use #:cl #:alexandria #:consfigurator #:consfigurator.property.disk)
   (:local-nicknames (#:os        #:consfigurator.property.os)
+                    (#:cmd       #:consfigurator.property.cmd)
                     (#:file      #:consfigurator.property.file)
                     (#:chroot    #:consfigurator.property.chroot)
+                    (#:mount     #:consfigurator.property.mount)
                     (#:fstab     #:consfigurator.property.fstab)
+                    (#:reboot    #:consfigurator.property.reboot)
                     (#:crypttab  #:consfigurator.property.crypttab))
   (:export #:install-bootloader-propspec
            #:install-bootloader-binaries-propspec
            #:chroot-installed-to-volumes
-           #:bootloader-binaries-installed))
+           #:bootloader-binaries-installed
+           #:bootloaders-installed
+           #:cleanly-installed-once))
 
 (defpackage :consfigurator.property.grub
   (:use #:cl #:alexandria #:consfigurator
@@ -554,7 +654,7 @@
            #:u-boot-installed-rockchip))
 
 (defpackage :consfigurator.property.hostname
-  (:use #:cl #:consfigurator)
+  (:use #:cl #:alexandria #:consfigurator)
   (:local-nicknames (#:cmd       #:consfigurator.property.cmd)
                     (#:container #:consfigurator.property.container)
                     (#:file      #:consfigurator.property.file))
@@ -564,10 +664,15 @@
            #:search-configured))
 
 (defpackage :consfigurator.property.network
-  (:use #:cl #:consfigurator)
+  (:use #:cl #:alexandria #:consfigurator)
   (:local-nicknames (#:os        #:consfigurator.property.os)
                     (#:file      #:consfigurator.property.file))
-  (:export #:static))
+  (:export #:aliases
+           #:ipv4
+           #:ipv6
+           #:clean-/etc/network/interfaces
+           #:static
+           #:preserve-static-once))
 
 (defpackage :consfigurator.property.libvirt
   (:use #:cl #:alexandria #:consfigurator)
@@ -581,11 +686,144 @@
            #:default-network-autostarted
            #:defined
            #:started
+           #:destroyed
            #:when-started
            #:kvm-boots-chroot-for
            #:kvm-boots-chroot-for.
            #:kvm-boots-chroot
-           #:kvm-boots-chroot.))
+           #:kvm-boots-chroot.
+           #:virsh-get-columns))
+
+(defpackage :consfigurator.property.ccache
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:os        #:consfigurator.property.os)
+                    (#:file      #:consfigurator.property.file)
+                    (#:apt       #:consfigurator.property.apt))
+  (:export #:installed
+           #:has-limits
+           #:group-cache))
+
+(defpackage :consfigurator.property.schroot
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:os        #:consfigurator.property.os)
+                    (#:file      #:consfigurator.property.file)
+                    (#:apt       #:consfigurator.property.apt))
+  (:export #:installed
+           #:uses-overlays
+           #:overlays-in-tmpfs))
+
+(defpackage :consfigurator.property.sbuild
+  (:use #:cl #:alexandria #:consfigurator)
+  (:local-nicknames (#:os        #:consfigurator.property.os)
+                    (#:file      #:consfigurator.property.file)
+                    (#:chroot    #:consfigurator.property.chroot)
+                    (#:user      #:consfigurator.property.user)
+                    (#:apt       #:consfigurator.property.apt)
+                    (#:ccache    #:consfigurator.property.ccache)
+                    (#:schroot   #:consfigurator.property.schroot)
+                    (#:periodic  #:consfigurator.property.periodic))
+  (:export #:installed
+           #:usable-by
+           #:built
+           #:built.
+           #:standard-debian-schroot))
+
+(defpackage :consfigurator.property.postfix
+  (:use #:cl #:alexandria #:consfigurator)
+  (:local-nicknames (#:cmd       #:consfigurator.property.cmd)
+                    (#:service   #:consfigurator.property.service)
+                    (#:apt       #:consfigurator.property.apt)
+                    (#:os        #:consfigurator.property.os)
+                    (#:file      #:consfigurator.property.file))
+  (:export #:installed
+           #:reloaded
+           #:main-configured
+           #:mapped-file))
+
+(defpackage :consfigurator.property.cron
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:re        #:cl-ppcre)
+                    (#:service   #:consfigurator.property.service)
+                    (#:apt       #:consfigurator.property.apt)
+                    (#:os        #:consfigurator.property.os)
+                    (#:file      #:consfigurator.property.file))
+  (:export #:system-job
+           #:nice-system-job
+           #:runs-consfigurator
+           #:user-crontab))
+
+(defpackage :consfigurator.property.lets-encrypt
+  (:use #:cl #:alexandria #:consfigurator)
+  (:local-nicknames (#:apt       #:consfigurator.property.apt)
+                    (#:os        #:consfigurator.property.os))
+  (:export #:installed
+           #:agree-tos
+           #:certificate-obtained
+           #:fullchain-for
+           #:chain-for
+           #:certificate-for
+           #:privkey-for))
+
+(defpackage :consfigurator.property.apache
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:service      #:consfigurator.property.service)
+                    (#:apt          #:consfigurator.property.apt)
+                    (#:os           #:consfigurator.property.os)
+                    (#:file         #:consfigurator.property.file)
+                    (#:network      #:consfigurator.property.network)
+                    (#:lets-encrypt #:consfigurator.property.lets-encrypt))
+  (:export #:installed
+           #:reloaded
+           #:mod-enabled
+           #:conf-enabled
+           #:conf-available
+           #:site-enabled
+           #:site-available
+           #:https-vhost))
+
+(defpackage :consfigurator.property.systemd
+  (:use #:cl #:consfigurator)
+  (:export #:started
+           #:stopped
+           #:enabled
+           #:disabled
+           #:masked
+           #:lingering-enabled))
+
+(defpackage :consfigurator.property.firewalld
+  (:use #:cl #:anaphora #:alexandria #:consfigurator)
+  (:local-nicknames (#:cmd          #:consfigurator.property.cmd)
+                    (#:file         #:consfigurator.property.file)
+                    (#:apt          #:consfigurator.property.apt)
+                    (#:os           #:consfigurator.property.os))
+  (:export #:installed
+           #:service
+           #:policy
+           #:zone
+           #:has-zone
+           #:zone-target
+           #:default-route-zoned-once
+           #:zone-has-interface
+           #:zone-has-service
+           #:zone-masquerade
+           #:zone-rich-rule
+           #:zone-direct-rule
+           #:default-zone))
+
+(defpackage :consfigurator.property.timezone
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:file         #:consfigurator.property.file)
+                    (#:apt          #:consfigurator.property.apt)
+                    (#:os           #:consfigurator.property.os))
+  (:export #:configured
+           #:configured-from-parent))
+
+(defpackage :consfigurator.property.swap
+  (:use #:cl #:consfigurator)
+  (:local-nicknames (#:cmd          #:consfigurator.property.cmd)
+                    (#:fstab        #:consfigurator.property.fstab)
+                    (#:os           #:consfigurator.property.os))
+  (:export #:has-swap-file))
 
 (defpackage :consfigurator.connection.local
   (:use #:cl #:consfigurator #:alexandria)
@@ -596,10 +834,9 @@
   (:export #:shell-wrap-connection #:connection-shell-wrap))
 
 (defpackage :consfigurator.connection.fork
-  (:use #:cl #:consfigurator #:consfigurator.connection.local)
+  (:use #:cl #:alexandria #:consfigurator #:consfigurator.connection.local)
   (:export #:fork-connection
-           #:post-fork
-	   #:can-probably-fork))
+           #:post-fork))
 
 (defpackage :consfigurator.connection.rehome
   (:use #:cl #:consfigurator #:consfigurator.connection.fork)
@@ -615,12 +852,14 @@
 (defpackage :consfigurator.connection.ssh
   (:use #:cl
         #:consfigurator
+        #:anaphora
         #:alexandria
         #:consfigurator.connection.shell-wrap))
 
 (defpackage :consfigurator.connection.sudo
   (:use #:cl
         #:consfigurator
+        #:anaphora
         #:alexandria
         #:consfigurator.connection.shell-wrap))
 
@@ -631,13 +870,15 @@
 
 (defpackage :consfigurator.connection.chroot
   (:use #:cl
+        #:anaphora
         #:alexandria
 	#:consfigurator
 	#:consfigurator.connection.fork
         #:consfigurator.connection.rehome
         #:consfigurator.connection.shell-wrap
 	#:cffi)
-  (:local-nicknames (#:disk      #:consfigurator.property.disk)))
+  (:local-nicknames (#:disk      #:consfigurator.property.disk)
+                    (#:mount     #:consfigurator.property.mount)))
 
 (defpackage :consfigurator.connection.setuid
   (:use #:cl
@@ -649,7 +890,7 @@
 		    (#:user #:consfigurator.property.user)))
 
 (defpackage :consfigurator.data.asdf
-  (:use #:cl #:consfigurator))
+  (:use #:cl #:alexandria #:consfigurator))
 
 (defpackage :consfigurator.data.pgp
   (:use #:cl #:consfigurator #:alexandria)
@@ -664,3 +905,6 @@
 (defpackage :consfigurator.data.ssh-askpass
   (:use #:cl #:alexandria #:consfigurator)
   (:local-nicknames (#:re   #:cl-ppcre)))
+
+(defpackage :consfigurator.data.local-file
+  (:use #:cl #:consfigurator))
