@@ -46,7 +46,7 @@
   (when apply
     (setf (get sym 'papply) apply))
   (when unapply
-    (setf (get sym 'unapply) unapply))
+    (setf (get sym 'punapply) unapply))
   sym)
 
 (defun isprop (prop)
@@ -63,9 +63,18 @@
       (get (car propapp) 'ptype)
       :posix))
 
+(defun propappargs (propapp)
+  (if (and (listp (cadr propapp)) (member :orig-args (cadr propapp)))
+      (getf (cadr propapp) :orig-args)
+      (cdr propapp)))
+
 (defun collapse-types (&rest lists)
   (if (member :posix (flatten lists)) :posix :lisp))
 
+(defun collapse-propapp-types (&rest lists)
+  (if (member :posix (mapcan (curry #'mapcar #'propapptype) lists))
+      :posix :lisp))
+
 (defun propdesc (prop &rest args)
   (apply (get prop 'desc #'noop) args))
 
@@ -96,10 +105,9 @@
 
 (defun propapply (prop &rest args)
   (with-some-errors-are-failed-change
-    (let ((check (get prop 'check)))
-      (if (and check (apply check args))
-          :no-change
-          (apply (get prop 'papply (constantly :no-change)) args)))))
+    (if (aand (get prop 'check) (apply it args))
+        :no-change
+        (apply (get prop 'papply (constantly :no-change)) args))))
 
 (defun propappapply (propapp)
   (if propapp
@@ -109,8 +117,8 @@
 (defun propunapply (prop &rest args)
   (with-some-errors-are-failed-change
     (let ((check (get prop 'check))
-          (apply (get prop 'apply))
-          (unapply (get prop 'unapply)))
+          (apply (get prop 'papply))
+          (unapply (get prop 'punapply)))
       ;; Only fail if there's no :UNAPPLY when there is an :APPLY, because
       ;; that is the case in which we can't do what was requested.  If there
       ;; is no :APPLY then we can infer that there is nothing on the host to
@@ -133,22 +141,10 @@
 (defvar *known-properties* nil
   "All properties whose definitions have been loaded.")
 
-(defvar *known-property-macrolets* nil
-  "Macro definitions for all known properties as used in MAP-PROPSPEC-PROPAPPS.
-
-This variable exists just to avoid consing these forms over and over again;
-see MAP-PROPSPEC-PROPAPPS for how they are used.")
-
 (defun record-known-property (psym)
   (unless (get psym 'isprop)
     (setf (get psym 'isprop) t)
-    (push psym *known-properties*)
-    (push `(,psym (&rest args)
-                  (let ((gensym (gensym)))
-                    (push (list* gensym ',psym args)
-                          *replaced-propapps*)
-                    gensym))
-          *known-property-macrolets*)))
+    (push psym *known-properties*)))
 
 (defun dump-properties-for-emacs (from to)
   (let ((put-forms
@@ -271,16 +267,17 @@ through unmodified, so supplied-p information is preserved."
                 (cons ',name (cdr ,whole)))))))))
 
 (defmacro define-property-defining-macro
-    (mname (typev lambdav slotsv formsv) &body mbody)
+    (mname (typev lambdav slotsv formsv &optional (namev (gensym))) &body mbody)
   "Define macro MNAME which be used to define properties, and which works by
 parsing FORMSV and pushing SETPROP keyword argument pairs to plist SLOTSV."
   (multiple-value-bind (mforms mdeclarations mdocstring)
       (parse-body mbody :documentation t)
     (declare (ignore mdeclarations))
-    (with-gensyms (name body declarations)
-      `(defmacro ,mname (,name ,typev ,lambdav &body ,body)
+    (with-gensyms (body declarations)
+      `(defmacro ,mname (,namev ,typev ,lambdav &body ,body)
          ,@(and mdocstring `(,mdocstring))
-         (let ((,slotsv (list :type ,typev :lambda `',,lambdav)))
+         (let ((programmatic-warning t)
+               (,slotsv (list :type ,typev :lambda `',,lambdav)))
            (multiple-value-bind (,formsv ,declarations)
                (parse-body ,body :documentation t)
              (when (> (length ,declarations) 1)
@@ -291,10 +288,10 @@ parsing FORMSV and pushing SETPROP keyword argument pairs to plist SLOTSV."
 					  (cdar ,declarations)))))
                `(progn
                   (eval-when (:compile-toplevel :load-toplevel :execute)
-                    (record-known-property ',,name))
-                  (store-indentation-info-for-emacs ',,name ',,lambdav ,indent)
-                  (setprop ',,name ,@,slotsv)
-                  (define-dotted-property-macro ,,name ,,lambdav)
+                    (record-known-property ',,namev))
+                  (store-indentation-info-for-emacs ',,namev ',,lambdav ,indent)
+                  (setprop ',,namev ,@,slotsv)
+                  (define-dotted-property-macro ,,namev ,,lambdav)
                   ;; Now prepare a DEFUN for the property, to enable calling
                   ;; it programmatically within the :APPLY and :UNAPPLY
                   ;; routines of other properties.  This can lead to clearer
@@ -302,21 +299,31 @@ parsing FORMSV and pushing SETPROP keyword argument pairs to plist SLOTSV."
                   ;; things like installing packages.
                   ,@(and
                      (getf ,slotsv :apply)
-                     `((defun-with-args ,,name args ,,lambdav
+                     `((defun-with-args ,,namev args ,,lambdav
+                         (unless *connection*
+                           (simple-program-error
+"Attempt to programmatically apply property with no connection established.
+
+Common causes of this error are (i) missing the '.' to enable the dotted
+propapp rules; and (ii) inadvertently calling the contents of a property's
+function cell, instead of constructing a propapp, within DEFPROPSPEC."))
                          ;; Properties with :HOSTATTRS subroutines which set
                          ;; new hostattrs should not be used programmatically
                          ;; in this way, so issue a warning.
-                         ,@(and (getf ,slotsv :hostattrs)
-                                '((programmatic-apply-hostattrs)))
-                         (consfigure (cons ',,name args)))))))))))))
+                         ,@(and programmatic-warning
+                                (getf ,slotsv :hostattrs)
+                                `((warn 'programmatic-apply-hostattrs
+                                        :property ',,namev)))
+                         (consfigure (cons ',,namev args)))))))))))))
 
-(define-condition programmatic-apply-hostattrs (simple-warning) ())
-
-(defun programmatic-apply-hostattrs ()
-  (warn 'programmatic-apply-hostattrs
-        :format-control
-        "Calling property which has :HOSTATTRS subroutine programmatically.
-Use DEFPROPLIST/DEFPROPSPEC to avoid trouble."))
+(define-condition programmatic-apply-hostattrs (warning)
+  ((property :initarg :property))
+  (:report (lambda (condition stream)
+             (format stream "Calling property ~S,
+which has :HOSTATTRS subroutine, programmatically.  Use DEFPROPLIST/DEFPROPSPEC
+to avoid trouble.  Use IGNORING-HOSTATTRS to muffle this warning if
+~:*~S does not push any new hostattrs."
+                     (slot-value condition 'property)))))
 
 (defmacro ignoring-hostattrs (form)
   "Where FORM is a programmatic call to a property which has a :HOSTATTRS
@@ -331,24 +338,24 @@ subroutine does not push any new hostattrs."
 
 ;; supported ways to write properties are DEFPROP, DEFPROPSPEC and DEFPROPLIST
 
-(define-property-defining-macro defprop (type lambda slots forms)
+(define-property-defining-macro defprop (type lambda slots forms name)
   "Define a property by providing code for its subroutines."
   (loop for form in forms
         if (keywordp (car form))
           do (setf (getf slots (car form)) (cdr form)))
   (loop for kw in '(:desc :preprocess :hostattrs :check :apply :unapply)
         do (if-let ((slot (getf slots kw)))
-             (setf (getf slots kw)
-                   ;; TODO wrap a BLOCK around ,@slot with the property name,
-                   ;; so we can return from it
-                   `(lambda ,lambda
-                      ,@(and (member kw '(:desc :hostattrs))
-                             `((declare
-                                (ignorable
-                                 ,@(ordinary-ll-variable-names
-                                    (ordinary-ll-without-&aux lambda)
-                                    :include-supplied-p t)))))
-                      ,@slot)))))
+             (multiple-value-bind (forms declarations) (parse-body slot)
+               (setf (getf slots kw)
+                     `(lambda ,lambda
+                        ,@(and (member kw '(:desc :hostattrs))
+                               `((declare
+                                  (ignorable
+                                   ,@(ordinary-ll-variable-names
+                                      (ordinary-ll-without-&aux lambda)
+                                      :include-supplied-p t)))))
+                        ,@declarations
+                        (block ,name ,@forms)))))))
 
 (define-property-defining-macro defpropspec (type lambda slots forms)
   "Define a property which constructs, evaluates and applies a propspec.
@@ -365,13 +372,14 @@ subroutines at the right time.
 If the first element of the body is a string, it will be considered a
 docstring for the resulting property.  If the first element of the body after
 any such string is a list beginning with :DESC, the remainder will be used as
-the :DESC subroutine for the resulting property, like DEFPROP.  Supplying a
-:CHECK subroutine in the same way is also supported.  Otherwise, the body
-defines a function of the arguments specified by the lambda list which returns
-the property application specification expression to be evaluated and applied.
-It should be a pure function aside from retrieving hostattrs (as set by other
-properties applied to the hosts to which the resulting property is applied,
-not as set by the properties in the returned propspec).
+the :DESC subroutine for the resulting property, like DEFPROP.  Supplying
+:CHECK and :HOSTATTRS subroutines in the same way is also supported.
+Otherwise, the body defines a function of the arguments specified by the
+lambda list which returns the property application specification expression to
+be evaluated and applied.  It should be a pure function aside from retrieving
+hostattrs (as set by other properties applied to the hosts to which the
+resulting property is applied, not as set by the properties in the returned
+propspec).
 
 Macro property combinators should be usable in the normal way in the body, but
 some other macros commonly used in DEFHOST and DEFPROPLIST forms will not work
@@ -403,15 +411,19 @@ You can usually use DEFPROPLIST instead of DEFPROPSPEC, which see."
   (loop while (and (listp (car forms)) (keywordp (caar forms)))
 	do (setf (getf slots (caar forms))
 		 `(lambda (plist)
-		    (destructuring-bind ,lambda (getf plist :orig-args)
-                      ,@(and (member (caar forms) '(:desc :hostattrs))
-                             `((declare
-                                (ignorable
-                                 ,@(ordinary-ll-variable-names
-                                    lambda :include-supplied-p t)))))
-		      ,@(cdr (pop forms))))))
+                    (with-*host*-*consfig*
+		      (destructuring-bind ,lambda (getf plist :orig-args)
+                        ,@(and (member (caar forms) '(:desc :hostattrs))
+                               `((declare
+                                  (ignorable
+                                   ,@(ordinary-ll-variable-names
+                                      lambda :include-supplied-p t)))))
+		        ,@(cdr (pop forms)))))))
+  (unless (getf slots :hostattrs)
+    (setq programmatic-warning nil))
   (setf (getf slots :hostattrs)
         `(lambda (plist)
+           ,@(cddr (getf slots :hostattrs))
            (let ((propspec (with-*host*-*consfig*
                                (preprocess-propspec
                                 (make-propspec
@@ -430,7 +442,7 @@ If the first element of PROPERTIES is a string, it will be considered a
 docstring for the resulting property.  If the first element of PROPERTIES
 after any such string is a list beginning with :DESC, the remainder will be
 used as the :DESC subroutine for the resulting property, like DEFPROP.
-Supplying a :CHECK subroutine in the same way is also supported.
+Supplying :CHECK and :HOSTATTRS subroutines in the same way is also supported.
 
 Otherwise, the body should not contain any references to variables other than
 those in LAMBDA.  LAMBDA is an ordinary lambda list, so you can use &AUX
@@ -446,17 +458,17 @@ You will usually be able to use DEFPROPLIST instead of DEFPROPSPEC.  However,
 sometimes you will need to fall back on DEFPROPSPEC.  For example, an
 unevaluated property application specification cannot express passing values
 other than constant values and propapps to property combinators."
-  (let ((propspec
-          (loop for remaining on properties
-                for car = (car remaining)
-                if (or (stringp car)
-                       (and (listp car) (member (car car)
-						'(:desc :check declare))))
-                  collect car into begin
-                else
-                  return (nreverse
-                          (cons `(props eseqprops ,@remaining) begin)))))
-    `(defpropspec ,name ,type ,lambda ,@propspec)))
+  (alet (loop for remaining on properties
+              for car = (car remaining)
+              if (or (stringp car)
+                     (and (listp car)
+                          (member (car car)
+                                  '(:desc :check :hostattrs declare))))
+                collect car into begin
+              else
+                return (nreverse
+                        (cons `(props eseqprops ,@remaining) begin)))
+    `(defpropspec ,name ,type ,lambda ,@it)))
 
 
 ;;;; hostattrs in property subroutines
@@ -470,7 +482,8 @@ install an apt package but the host is FreeBSD.")
   "HOST value currently being preprocessed.
 Used by GET-HOSTATTRS to break infinite loops.")
 
-(defun get-hostattrs (k &optional (host *host*) &aux (host (ensure-host host)))
+(defun get-hostattrs
+    (k &optional host &aux (host (ensure-host (or host *host*))))
   "Retrieve the list of static informational attributes of type KEY.
 
 Called by property :HOSTATTRS, :APPLY and :UNAPPLY subroutines."
@@ -504,19 +517,24 @@ Called by property :HOSTATTRS subroutines."
   (setf (getf (slot-value *host* 'hostattrs) k)
         (append vs (get-hostattrs k))))
 
-(defun pushnew-hostattrs (k &rest vs)
-  "Push new static informational attributes VS of type KEY.
+(defun pushnew-hostattr (k v &key (test #'equal))
+  "Push new static informational attribute V of type K.
+TEST is passed on to PUSHNEW.  Called by property :HOSTATTRS subroutines."
+  (pushnew-hostattrs k (list v) :test test))
 
-Called by property :HOSTATTRS subroutines."
+(defun pushnew-hostattrs (k vs &key (test #'equal))
+  "Push new static informational attributes VS of type K.
+VS is a list of items.  TEST is passed on to PUSHNEW.  Called by property
+:HOSTATTRS subroutines."
   (dolist (v (reverse vs))
-    (pushnew v (getf (slot-value *host* 'hostattrs) k))))
+    (pushnew v (getf (slot-value *host* 'hostattrs) k) :test test)))
 
 (defun require-data (iden1 iden2)
-  "Wrapper around PUSH-HOSTATTRS to indicate that a piece of prerequisite data
+  "Wrapper around PUSHNEW-HOSTATTR to indicate that a piece of prerequisite data
 is needed to deploy a property.
 
 Called by property :HOSTATTRS subroutines."
-  (pushnew-hostattrs :data (cons iden1 iden2)))
+  (pushnew-hostattr :data (cons iden1 iden2)))
 
 (defun get-hostname (&optional (host *host*))
   "Get the hostname of HOST, defaulting to the host to which properties are
@@ -544,7 +562,7 @@ apply or unapply properties.")
 PATH already has the specified CONTENT and MODE."
   (if (and (remote-exists-p path)
            (multiple-value-bind (existing-mode existing-size)
-               (remote-file-mode-and-size path)
+               (remote-file-stats path)
              (and (or (not mode-supplied-p) (= mode existing-mode))
                   (and (>= (* 4 (length content)) existing-size)
                        (string= (readfile path) content)))))
@@ -570,11 +588,14 @@ PATH already has the specified CONTENT and MODE."
 
 ;; this is a safe parse of ls(1) output given its POSIX specification
 (defun ls-cksum (file)
-  (let ((ls (ignore-errors
-             (split-string (run :env '(:LOCALE "C") "ls" "-dlL" file))))
-        (cksum (ignore-errors (cksum file))))
-    (when (and ls cksum)
-      (list* (car ls) cksum (subseq ls 2 8)))))
+  (when-let* ((ls (ignore-errors
+                   (words (run :env '(:LC_ALL "C") "ls" "-dlL" file))))
+              (ls-car (car ls))
+              (ls-end (subseq ls 2 8)))
+    (if (char= #\d (elt ls-car 0))
+        (cons ls-car ls-end)
+        (let ((cksum (ignore-errors (cksum file))))
+          (and cksum (list* ls-car cksum ls-end))))))
 
 (defmacro with-change-if-changes-file ((file) &body forms)
   "Execute FORMS and yield :NO-CHANGE if FILE does not change.
@@ -586,7 +607,21 @@ changes in properties which will change the file but not the output of `ls
   (with-gensyms (before)
     `(let* ((,before (ls-cksum ,file))
             (result (progn ,@forms)))
-       (if (and ,before (equal ,before (ls-cksum ,file)))
+       (if (or (eql result :no-change)
+               (and ,before (equal ,before (ls-cksum ,file))))
+           :no-change result))))
+
+(defmacro with-change-if-changes-files ((&rest files) &body forms)
+  "Execute FORMS and yield :NO-CHANGE if none of FILES change.
+See WITH-CHANGE-IF-CHANGES-FILE docstring regarding the sense of 'change'."
+  (with-gensyms (filesg beforeg)
+    `(let* ((,filesg (list ,@files))
+            (,beforeg (mapcar #'ls-cksum ,filesg))
+            (result (progn ,@forms)))
+       (if (or (eql result :no-change)
+               (loop for file in ,filesg and before in ,beforeg
+                     always before
+                     always (equal before (ls-cksum file))))
            :no-change result))))
 
 (defmacro with-change-if-changes-file-content ((file) &body forms)
@@ -594,7 +629,8 @@ changes in properties which will change the file but not the output of `ls
   (with-gensyms (before)
     `(let* ((,before (ignore-errors (cksum ,file)))
             (result (progn ,@forms)))
-       (if (and ,before (eql ,before (cksum ,file)))
+       (if (or (eql result :no-change)
+               (and ,before (eql ,before (cksum ,file))))
            :no-change result))))
 
 (defmacro with-change-if-changes-file-content-or-mode ((file) &body forms)
@@ -603,8 +639,10 @@ afterwards."
   (with-gensyms (before)
     `(let* ((,before (ls-cksum ,file))
             (result (progn ,@forms)))
-       (let ((after (ls-cksum ,file)))
-         (if (and ,before
-                  (string= (car ,before) (car after) :start1 1 :start2 1)
-                  (eql (cadr ,before) (cadr after)))
-             :no-change result)))))
+       (if (equal result :no-change)
+           :no-change
+           (let ((after (ls-cksum ,file)))
+             (if (and ,before
+                      (string= (car ,before) (car after) :start1 1 :start2 1)
+                      (eql (cadr ,before) (cadr after)))
+                 :no-change result))))))
