@@ -26,26 +26,58 @@
    (if (test "-d" (merge-pathnames "debootstrap/"
                                    (ensure-directory-pathname root)))
        (progn (delete-remote-trees root) nil)
-       (remote-exists-p (merge-pathnames "usr/lib/os-release" root))))
+       (remote-exists-p (chroot-pathname "/usr/lib/os-release" root))))
   (:apply
-   (let* ((os (get-hostattrs-car :os host))
+   (destructuring-bind
+       (&key (apt.proxy (get-hostattrs-car :apt.proxy host))
+          (apt.mirror (get-hostattrs-car :apt.mirror host))
+        &allow-other-keys
+        &aux (os (get-hostattrs-car :os host))
           (args (list "debootstrap"
-                      (plist-to-cmd-args options)
+                      (plist-to-cmd-args
+                       (remove-from-plist options :apt.proxy :apt.mirror))
                       (strcat "--arch=" (os:debian-architecture os))
                       (os:debian-suite os)
                       root)))
-     (when-let ((proxy (get-hostattrs-car :apt.proxy)))
-       (setq args (list* :env (list :http_proxy proxy) args)))
-     (when-let ((mirror (get-hostattrs-car :apt.mirror)))
-       (nconcf args (list mirror)))
+       options
+
+     ;; In the case where the chroot arch is not equal to the host arch, we
+     ;; could execute arch-test(1) here to confirm the architecture is
+     ;; executable by the running kernel; we'd add arch-test alongside
+     ;; qemu-user-static in %OS-BOOTSTRAPPER-INSTALLED.  Or possibly we only
+     ;; try to execute arch-test(1) when we find it's already on PATH.
+
+     (when apt.proxy
+       (setq args (list* :env (list :http_proxy apt.proxy) args)))
+     (when apt.mirror
+       (nconcf args (list apt.mirror)))
      (apply #'run args))))
 
+(defprop %debootstrap-manually-installed :posix ()
+  (:check (zerop (mrun :for-exit "command" "-v" "debootstrap")))
+  (:apply
+   (failed-change "Don't know how to install debootstrap(8) manually.")))
+
 (defpropspec %os-bootstrapper-installed :posix (host)
-  (:desc (declare (ignore host)) "OS bootstrapper installed")
-  `(os:host-etypecase ,host
-     (debian
-      (os:etypecase
-        (debianlike (apt:installed "debootstrap" "qemu-user-static"))))))
+  (:desc "OS bootstrapper installed")
+  (let ((host (preprocess-host host)))
+    `(os:host-etypecase ,host
+       (debian
+        ;; Have %DEBOOTSTRAP-MANUALLY-INSTALLED like this to enable installing
+        ;; Debian on arbitrary unixes, where Consfigurator doesn't know how to
+        ;; install packages, but the user has manually ensured that
+        ;; debootstrap(8) is on PATH.  However, we don't have such an escape
+        ;; hatch for the case where the architectures do not match because
+        ;; ensuring that debootstrap(8) will be able to bootstrap a foreign
+        ;; arch is more involved.
+        (os:typecase
+          (debianlike (apt:installed "debootstrap"))
+          (t (%debootstrap-manually-installed)))
+        ,@(and (not (call-with-os
+                     #'os:supports-arch-p
+                     (os:linux-architecture (get-hostattrs-car :os host))))
+               '((os:etypecase
+                   (debianlike (apt:installed "qemu-user-static")))))))))
 
 (defpropspec %os-bootstrapped :posix (options root host)
   "Bootstrap OS into ROOT, e.g. with debootstrap(1)."
@@ -80,7 +112,8 @@ starting services in the chroot, and set up access to parent hostattrs."
   (:desc #?"Subdeployment of ${root}")
   (consfigurator:deploys
    `((:chroot :into ,root))
-   (%make-child-host (replace-propspec-into-host host properties))))
+   (%make-child-host
+    (replace-propspec-into-host (ensure-host host) properties))))
 
 (defproplist os-bootstrapped-for :lisp
     (options root host &optional additional-properties
@@ -93,9 +126,11 @@ OPTIONS is a plist of values to pass to the OS-specific bootstrapping property."
   (:desc
    (declare (ignore options))
    #?"Built chroot for ${(get-hostname child-host*)} @ ${root}")
-  (%os-bootstrapper-installed child-host*)
-  (%os-bootstrapped options root child-host*)
-  (consfigurator:deploys `((:chroot :into ,root)) child-host))
+  (with-unapply
+    (%os-bootstrapper-installed child-host*)
+    (%os-bootstrapped options root child-host*)
+    (consfigurator:deploys `((:chroot :into ,root)) child-host)
+    :unapply (mount:unmounted-below-and-removed root)))
 
 (defproplist os-bootstrapped :lisp (options root properties)
   "Bootstrap an OS into ROOT and apply PROPERTIES.
