@@ -42,8 +42,25 @@ CONTENT can be a list of lines or a single string."
                    (string (format nil "~A~&" content)))
                  (and mode-supplied-p `(:mode ,mode)))))
 
+(defpropspec exists-with-content :posix
+    (path content &key (mode nil mode-supplied-p))
+  "Like FILE:HAS-CONTENT, but unapplicable, where unapplying means deleting
+PATH, rather than restoring any old content it might have had.
+
+The semantics of this property are that if it is applied, the file has content
+CONTENT, and if it is not applied to a host, the file does not exist.  Thus,
+it should be used to create new files on the host, not to replace the contents
+of existing files, such as those installed by operating system packages.  For
+replacing the contents of existing files, prefer FILE:HAS-CONTENT."
+  (declare (indent 1))
+  (:desc #?"${path} exists with defined content")
+  `(with-unapply
+     (has-content ,path ,content ,@(and mode-supplied-p `(:mode ,mode)))
+     :unapply (does-not-exist ,path)))
+
 (defprop contains-lines :posix (path &rest lines)
   "Ensure there is a file at PATH containing each of LINES once."
+  (declare (indent 1))
   (:apply
    (let ((new-lines (copy-list (ensure-cons lines)))
          (existing-lines (and (remote-exists-p path)
@@ -61,6 +78,17 @@ CONTENT can be a list of lines or a single string."
     path
     (curry #'remove-if (lambda (l) (member l lines :test #'string=))))))
 
+(defprop lacks-lines-matching :posix (path &rest patterns)
+  "If there is a file a PATH, ensure it does not contain any lines matching
+any of the regular expressions PATTERNS."
+  (:apply
+   (let ((scanners (mapcar #'re:create-scanner patterns)))
+     (map-file-lines
+      path (lambda (lines)
+             (loop for line in lines
+                   unless (loop for s in scanners thereis (re:scan s line))
+                     collect line))))))
+
 (defprop has-mode :posix (path mode)
   "Ensure that a file has a particular numeric mode."
   (:desc (format nil "~A has mode ~O" path mode))
@@ -68,6 +96,19 @@ CONTENT can be a list of lines or a single string."
    (with-change-if-changes-file (path)
      (mrun (format nil "chmod ~O ~A" mode path)))))
 
+(defprop has-ownership :posix (path &key user group)
+  "Ensure that a file has particular ownership and group ownership."
+  (:desc (format nil "~A has~:[~; owner ~:*~A~]~:[~;~2:*~:[~;,~] group ~A~]"
+                 path user group))
+  (:hostattrs
+   (unless (or user group)
+     (inapplicable-property "Not enough arguments.")))
+  (:apply
+   (with-change-if-changes-file (path)
+     (if user
+         (mrun "chown" "-h" (format nil "~A~:[~;:~:*~A~]" user group) path)
+         (mrun "chgrp" "-h" group path)))))
+
 (defprop does-not-exist :posix (&rest paths)
   "Ensure that files do not exist."
   (:desc (if (cdr paths)
@@ -76,6 +117,21 @@ CONTENT can be a list of lines or a single string."
   (:check (not (apply #'remote-exists-p paths)))
   (:apply (mrun "rm" "-f" paths)))
 
+(defprop directory-does-not-exist :posix
+    (&rest directories
+           ;; Ensure that there's a trailing slash at the end of each
+           ;; namestring, such that if a regular file of the same name as the
+           ;; directory exists, 'rm -rf' will not delete it.
+           &aux (directories
+                 (mapcar (compose #'ensure-trailing-slash #'unix-namestring)
+                         directories)))
+  "Recursively ensure that DIRECTORIES do not exist."
+  (:desc (if (cdr directories)
+             #?"@{directories} do not exist"
+             #?"${(car directories)} does not exist"))
+  (:check (not (apply #'remote-exists-p directories)))
+  (:apply (mrun "rm" "-rf" directories)))
+
 (defprop data-uploaded :posix (iden1 iden2 destination)
   (:hostattrs
    (declare (ignore destination))
@@ -84,11 +140,13 @@ CONTENT can be a list of lines or a single string."
    (containing-directory-exists destination)
    (maybe-writefile-data destination iden1 iden2)))
 
-(defprop host-data-uploaded :posix (destination)
-  (:hostattrs
-   (require-data (get-hostname) destination))
-  (:apply
-   (data-uploaded (get-hostname) destination destination)))
+(defproplist host-data-uploaded :posix
+    (destination
+     ;; Require an absolute path because we don't know the remote home
+     ;; directory at hostattrs time, so can't resolve it ourselves.
+     &aux (destination (unix-namestring
+                        (ensure-pathname destination :want-absolute t))))
+  (data-uploaded (get-hostname) destination destination))
 
 (defprop secret-uploaded :posix (iden1 iden2 destination)
   (:hostattrs
@@ -98,12 +156,17 @@ CONTENT can be a list of lines or a single string."
    (containing-directory-exists destination)
    (maybe-writefile-data destination iden1 iden2 :mode #o600)))
 
-(defprop host-secret-uploaded :posix
-    (destination &aux (destination (unix-namestring destination)))
-  (:hostattrs
-   (require-data (get-hostname) destination))
-  (:apply
-   (secret-uploaded (get-hostname) destination destination)))
+(defproplist host-secret-uploaded :posix
+    (destination
+     ;; Require an absolute path like with HOST-DATA-UPLOADED.
+     &aux (destination (unix-namestring
+                        (ensure-pathname destination :want-absolute t))))
+  (secret-uploaded (get-hostname) destination destination))
+
+(defprop data-cache-purged :posix ()
+  "Ensure that any prerequisite data cached in the remote home directory is removed."
+  (:desc "Consfigurator data cache cleaned")
+  (:apply (directory-does-not-exist (get-remote-data-cache-dir))))
 
 (defprop regex-replaced-lines :posix (file regex replace)
   "Like s/REGEX/REPLACE/ on the lines of FILE.
@@ -134,14 +197,19 @@ Uses CL-PPCRE:REGEX-REPLACE, which see for the syntax of REPLACE."
    ;; likewise assume it was already there
    :no-change))
 
-;; readlink(1) is not POSIX
+;; readlink(1) is not POSIX.  This is a safe parse of ls(1) output given its
+;; POSIX specification.
 (defun remote-link-target (symlink)
-  (loop with s = (stripln (run :env '(:LOCALE "POSIX") "ls" "-ld" symlink))
-        with found = 0
+  (loop with s = (stripln (run :env '(:LC_ALL "C") "ls" "-ld" symlink))
+        and found = 0 and just-found
         for i from 0 below (length s)
-        when (char= (elt s i) #\Space)
-          do (incf found)
-        when (>= found 9)
+
+        if (char= (elt s i) #\Space)
+          unless just-found do (incf found) (setq just-found t)
+            end
+        else do (setq just-found nil)
+
+        when (>= found 8)
           return (subseq s (+ (length symlink) i 5))))
 
 (defprop symlinked :posix (&key from to)
@@ -258,7 +326,7 @@ Other arguments:
                        collect (list commentedp current-section k v)
                      else collect line))
              (mapped (funcall map unmapped)))
-        (loop with current-section
+        (loop with current-section and collectedp
               for line in mapped
               for line-section = (etypecase line
                                    (cons (cadr line))
@@ -267,7 +335,7 @@ Other arguments:
               if (and (listp line)
                       line-section
                       (not (string= line-section current-section)))
-                collect ""
+                when collectedp collect "" end
                 and collect (funcall new-section line-section)
                 and do (setq current-section line-section)
               else if (and (stringp line)
@@ -281,7 +349,8 @@ Other arguments:
                             (declare (ignore sec))
                             (when commentedp (princ new-comment s))
                             (princ (funcall new-kv k v) s)))
-              else collect line)))))
+              else collect line
+              do (setq collectedp t))))))
 
 (defun simple-conf-update (file pairs &rest args)
   (let ((keys (make-hash-table :test #'equal)))
@@ -342,7 +411,7 @@ commented out; the first commented or uncommented line for each key will be
 uncommented and used to set the value, if it exists."
   (:desc (format nil "~A has ~{~A = ~A~^, ~}" file pairs))
   (:apply (simple-conf-update file pairs
-                              :parse-kv #?/^(\S+) = (.+)/
+                              :parse-kv #?/^([^\s=]+)\s?=\s?(.*)/
                               :new-kv (lambda (k v) #?"${k} = ${v}"))))
 
 (defprop contains-conf-shell :posix (file &rest pairs)
@@ -360,7 +429,7 @@ uncommented and used to set the value, if it exists."
                               ;; include quoting as part of the value so we
                               ;; don't end up substituting double quotation
                               ;; marks for single, or similar
-                              :parse-kv #?/^(\S+)\s?=\s?(.*)/
+                              :parse-kv #?/^([^\s=]+)\s?=\s?(.*)/
                               :new-kv (lambda (k v) #?"${k}=${v}"))))
 
 (defprop contains-ini-settings :posix (file &rest triples)
@@ -436,7 +505,15 @@ ENTRIES, using a simple merge procedure: existing lines of the file with the
 same value for the TARGETth field are updated to match the corresponding
 members of ENTRIES, except that if the SOURCEth field of the existing entry is
 not NO-SOURCE and the corresponding member of ENTRIES is STRING= to either
-NO-SOURCE or \"PLACEHOLDER\", use the existing field value."
+NO-SOURCE or \"PLACEHOLDER\", use the existing field value.
+
+Sort the lines to avoid certain possible failures.  For each pair of lines, if
+the TARGETth or the SOURCEth field of the first line is a path and a subpath
+of the TARGETth field of the second line, sort the second line earlier.
+Otherwise, try to avoid disturbing line order.  This avoids failures to mount
+because the filesystem containing the mount point is not mounted yet, and
+ensures that partitions containing things like swap files are mounted before
+an attempt is made to activate the swap, set up the bind mount, etc."
   (let ((unknown (list no-source "PLACEHOLDER"))
         (pending (make-hash-table :test #'equal)))
     (dolist (entry entries)
@@ -444,18 +521,36 @@ NO-SOURCE or \"PLACEHOLDER\", use the existing field value."
     (map-file-lines
      file
      (lambda (lines)
-       (loop for line in lines
-             for line-fields = (words line)
-             for line-source = (nth source line-fields)
-             and line-target = (nth target line-fields)
-             for entry = (when-let* ((entry (gethash line-target pending))
-                                     (fields (words entry)))
-                           (when (and (member (nth source fields)
-                                              unknown :test #'string=)
-                                      (not (string= line-source no-source)))
-                             (setf (nth source fields) line-source))
-                           (format nil "~{~A~^ ~}" fields))
-             if entry
-               collect it into accum and do (remhash line-target pending)
-             else collect line into accum
-             finally (return (nconc accum (hash-table-values pending))))))))
+       (stable-sort
+        (loop for line in lines
+              for line-fields = (words line)
+              for line-source = (nth source line-fields)
+              and line-target = (nth target line-fields)
+              for entry = (when-let* ((entry (gethash line-target pending))
+                                      (fields (words entry)))
+                            (when (and (member (nth source fields)
+                                               unknown :test #'string=)
+                                       (not (string= line-source no-source)))
+                              (setf (nth source fields) line-source))
+                            (format nil "~{~A~^ ~}" fields))
+              if entry
+                collect it into accum and do (remhash line-target pending)
+              else collect line into accum
+              finally (return (nconc accum (hash-table-values pending))))
+        (lambda (a b)
+          (flet ((subpathp (x y)
+                   (and (plusp (length x)) (plusp (length y))
+                        (char= #\/ (first-char x) (first-char y))
+                        (subpathp (ensure-pathname x)
+                                  (ensure-directory-pathname y)))))
+            (and
+             ;; If either line is blank treat as equal.
+             (plusp (length a)) (plusp (length b))
+             ;; If either line is a comment treat as equal.
+             (not (char= #\# (first-char a))) (not (char= #\# (first-char b)))
+             ;; Now compare the TARGETth and SOURCEth fields of B to the
+             ;; TARGETth field of A.
+             (let ((a (words a)) (b (words b)))
+               (let ((a-target (nth target a)))
+                 (or (subpathp (nth target b) a-target)
+                     (subpathp (nth source b) a-target))))))))))))
