@@ -41,7 +41,30 @@
   (values))
 
 (defun lines (text)
-  (split-string (stripln text) :separator '(#\Newline #\Return)))
+  (with-input-from-string (stream text)
+    (let (bolp buffer)
+      (flet ((reset ()
+               (setq bolp t
+                     buffer (make-array 0 :fill-pointer 0
+                                          :element-type 'character))))
+        ;; Split on either <CR>, <LF> or <CR><LF>; <LF><CR> would mean split
+        ;; with a blank line in between.  Drop a single trailing blank line.
+        (loop initially (reset)
+              for char = (read-char stream nil nil)
+              if char
+                if (member char '(#\Return #\Newline) :test #'char=)
+                  collect buffer
+                  and do (reset)
+                         (when (char= char #\Return)
+                           (when-let ((next (peek-char nil stream nil nil)))
+                             (when (char= next #\Newline)
+                               (read-char stream))))
+                else do (setq bolp nil)
+                        (vector-push-extend char buffer)
+                end
+              else
+                unless bolp collect buffer end
+                and do (loop-finish))))))
 
 (defun unlines (lines)
   (format nil "~{~A~%~}" lines))
@@ -165,13 +188,34 @@ one solution is to convert your property to a :LISP property."
   "Like PATHNAME-NAME but include any file extension."
   (and (pathname-name pathname)
        (namestring
-        (enough-pathname pathname (pathname-directory-pathname pathname)))))
+        (if (pathname-directory pathname)
+            (enough-pathname pathname (pathname-directory-pathname pathname))
+            pathname))))
+
+(defun directory-contents (pathname)
+  "Return the immediate contents of PATHNAME, a directory, without resolving
+symlinks.  Not suitable for use by :POSIX properties."
+  ;; On SBCL on Debian UIOP:*WILD-FILE-FOR-DIRECTORY* is #P"*.*".
+  (uiop:directory*
+   (merge-pathnames uiop:*wild-file-for-directory*
+                    (ensure-directory-pathname pathname))))
+
+(defun ensure-trailing-slash (namestring)
+  (if (string-suffix-p namestring "/")
+      namestring
+      (strcat namestring "/")))
 
 (defun drop-trailing-slash (namestring)
   (if (string-suffix-p namestring "/")
       (subseq namestring 0 (1- (length namestring)))
       namestring))
 
+(defun reinit-structlike (class &rest slots)
+  (loop with object = (allocate-instance (find-class class))
+        for (slot-name slot-value) on slots by #'cddr
+        do (setf (slot-value object slot-name) slot-value)
+        finally (return (reinitialize-instance object))))
+
 (defmacro quote-nonselfeval (x)
   (once-only (x)
     `(if (member (type-of ,x) '(cons symbol))
@@ -181,17 +225,16 @@ one solution is to convert your property to a :LISP property."
   "Define an implementation of PRINT-OBJECT for objects which are simple
 one-dimensional collections of values."
   `(defmethod print-object ((object ,class) stream)
-     (if *print-readably*
+     (if (and *print-readably* *read-eval*)
          (format
           stream "#.~S"
-          `(make-instance
+          `(reinit-structlike
             ',(type-of object)
             ;; Call CLASS-OF so that subclasses of CLASS are handled too.
             ,@(loop for slot in (closer-mop:class-slots (class-of object))
-                    for initargs = (closer-mop:slot-definition-initargs slot)
-                    and slot-name = (closer-mop:slot-definition-name slot)
+                    for slot-name = (closer-mop:slot-definition-name slot)
                     when (slot-boundp object slot-name)
-                      collect (car initargs)
+                      collect `',slot-name
                       and collect (quote-nonselfeval
                                    (slot-value object slot-name)))))
          (call-next-method))
@@ -209,9 +252,9 @@ one-dimensional collections of values."
   "Like UIOP:ESCAPE-SH-TOKEN, but also escape the empty string."
   (if (string= token "") (format s "\"\"") (uiop:escape-sh-token token s)))
 
-(defun escape-sh-command (token &optional s)
+(defun escape-sh-command (command &optional s)
   "Like UIOP:ESCAPE-SH-COMMAND, but also escape the empty string."
-  (uiop:escape-command token s 'escape-sh-token))
+  (uiop:escape-command command s 'escape-sh-token))
 
 (defun parse-username-from-id (output)
   "Where OUTPUT is the output of the id(1) command, extract the username."
@@ -252,6 +295,99 @@ expansion as a starting point for your own DEFPACKAGE form for your consfig."
     ;;   (push '(:use '#:cl '#:consfigurator) forms))
     `(defpackage ,name ,@forms)))
 
+(defmacro lambda-ignoring-args (&body body)
+  (multiple-value-bind (forms declarations) (parse-body body)
+    (with-gensyms (ignore)
+      `(lambda (&rest ,ignore)
+         (declare (ignore ,ignore) ,@declarations)
+         ,@forms))))
+
+(defun parse-cidr (address-with-suffix)
+  (destructuring-bind (address cidr)
+      (split-string address-with-suffix :separator "/")
+    (unless cidr
+      (simple-program-error "~A is not in CIDR notation."
+                            address-with-suffix))
+    (values
+     address
+     (loop with cidr = (parse-integer cidr)
+           with type = (if (or (> cidr 32) (find #\: address)) 6 4)
+           with block-digits = (if (= type 4) 8 16)
+           repeat (if (= type 4) 4 8)
+           for digits = (min cidr block-digits)
+           do (decf cidr digits)
+           collect (parse-integer
+                    (with-output-to-string (s)
+                      (loop repeat digits do (princ #\1 s))
+                      (loop repeat (- block-digits digits) do (princ #\0 s)))
+                    :radix 2)
+             into accum
+           finally (return (if (= type 4)
+                               (format nil "~{~D~^.~}" accum)
+                               (with-output-to-string (s)
+                                 (loop for blocks on accum
+                                       if (> (car blocks) 0)
+                                         do (format s "~X" (car blocks))
+                                         and if (cdr blocks) do (princ #\: s)
+                                               end
+                                       else do (princ #\: s)
+                                               (loop-finish)))))))))
+
+(defun system (&rest args)
+  "Simple wrapper around system(3)."
+  (foreign-funcall
+   "system" :string (if (cdr args)
+                        (escape-sh-command
+                         (loop for arg in args
+                               if (pathnamep arg)
+                                 collect (unix-namestring arg)
+                               else collect arg))
+                        (car args))
+   :int))
+
+(define-constant +alphanum+
+  "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
+  :test #'string=)
+
+(defun mkfifo ()
+  "Use mkfifo(3) to create a named pipe with a mkstemp(3)-like name."
+  (let* ((dir (drop-trailing-slash (or (getenv "TMPDIR") "/tmp")))
+         (dir-ls (run-program
+                  `("env" "LANG=C" "ls" "-lnd" ,dir) :output :string))
+         (prefix (strcat dir "/tmp.")))
+    (unless (and (char= #\d (char dir-ls 0))
+                 (char-equal #\t (char dir-ls 9))
+                 (zerop (parse-integer (caddr (words dir-ls)))))
+      (error "~A is not a root-owned dir with the sticky bit set." dir))
+    (flet ((mktemp ()
+             ;; We need to generate a temporary name.  We don't have to worry
+             ;; about race conditions as mkfifo(3) will fail if the file
+             ;; already exists.
+             (loop with result = (make-string (+ 6 (length prefix)))
+                     initially (setf (subseq result 0 (length prefix)) prefix)
+                   for i from (length prefix) below (length result)
+                   do (setf (char result i)
+                            (char +alphanum+ (random #.(length +alphanum+))))
+                   finally (return result)))
+           (mkfifo (temp)
+             (handler-case (nix:mkfifo temp #o600)
+               (serious-condition (c)
+                 (if (or (file-exists-p temp) (directory-exists-p temp))
+                     nil
+                     (signal c))))))
+      (loop with *random-state* = (make-random-state t)
+            repeat 3 for temp = (mktemp)
+            when (mkfifo temp) return (pathname temp)))))
+
+(defmacro with-mkfifos ((&rest mkfifos) &body forms)
+  `(let ,(loop for mkfifo in mkfifos collect `(,mkfifo (mkfifo)))
+     (unwind-protect (progn ,@forms)
+       ,@(loop for mkfifo in mkfifos collect `(delete-file ,mkfifo)))))
+
+(defun write-to-mkfifo (object fifo)
+  (with-standard-io-syntax
+    (write object :stream fifo) (terpri fifo) (finish-output fifo)))
+
 
 ;;;; Progress & debug printing
 
@@ -363,29 +499,103 @@ previous output."
 
 ;;;; Forking utilities
 
-(define-condition in-child-process () ())
+;;; Use implementation-specific fork(2) wrapper, and never fork(2) itself, to
+;;; allow the implementation to handle things like finaliser threads.  For all
+;;; other syscalls/libc & POSIX macros like WIFEXITED, use CFFI, via Osicat
+;;; when there's a wrapper available, for portability.
 
-(defmacro unwind-protect-in-parent (protected &body cleanup)
-  "Like UNWIND-PROTECT, but with a mechanism to cancel the execution of CLEANUP
-in child processes resulting from calls to fork(2) during the execution of
-PROTECTED.  This means that CLEANUP won't get executed on both sides of the
-fork, but only in the parent.
+(defun fork ()
+  ;; Normalise any other implementations such that we signal an error if
+  ;; fork(2) returns -1, so caller doesn't have to check for that.
+  #+sbcl (sb-posix:fork))
 
-For this to work, after fork(2), the child process must call
-CANCEL-UNWIND-PROTECT-IN-PARENT-CLEANUP, which will affect all enclosing uses
-of this macro."
-  (with-gensyms (cancelled)
-    `(let (,cancelled)
-       (unwind-protect
-            (handler-bind ((in-child-process
-                             (lambda (c) (setq ,cancelled t) (signal c))))
-              ,protected)
-         (unless ,cancelled ,@cleanup)))))
+(defmacro forked-progn (child-pid child-form &body parent-forms)
+  (with-gensyms (retval)
+    `(progn
+       #-(or sbcl) (error "Don't know how to safely fork(2) in this Lisp.")
+       (mapc-open-output-streams
+        #'force-output
+        *standard-output* *error-output* *debug-io* *terminal-io*)
+       (let ((,retval (fork)))
+         (if (zerop ,retval)
+             ;; We leave it to the caller to appropriately call CLOSE or
+             ;; CLEAR-INPUT on input streams shared with the parent, because
+             ;; at least SBCL's CLEAR-INPUT clears the OS buffer as well as
+             ;; Lisp's, potentially denying data to both sides of the fork.
+             ,child-form
+             (let ((,child-pid ,retval)) ,@parent-forms))))))
 
-(defun cancel-unwind-protect-in-parent-cleanup ()
-  "Cancel the CLEANUP forms in all enclosing uses of UNWIND-PROTECT-IN-PARENT.
-Should be called soon after fork(2) in child processes."
-  (signal 'in-child-process))
+(define-condition skipped-properties () ()
+  (:documentation
+   "There were failed changes, but instead of aborting, that particular property
+application was instead skipped over, either due to the semantics of a
+property combinator, or because the user elected to skip the property in the
+interactive debugger."))
+
+(defmacro with-deployment-report (&body forms)
+  (with-gensyms (failures)
+    `(let* (,failures
+            (result (handler-bind ((skipped-properties (lambda (c)
+                                                         (declare (ignore c))
+                                                         (setq ,failures t))))
+                      ,@forms)))
+       (inform
+        t
+        (cond
+          ((eql :no-change result)
+           "No changes were made.")
+          (,failures
+           "There were failures while attempting to apply some properties.")
+          (t
+           "Changes were made without any reported failures."))))))
+
+(defmacro with-backtrace-and-exit-code (&body forms)
+  (with-gensyms (failures)
+    `(let* (,failures
+            (result (handler-bind ((serious-condition
+                                     (lambda (c)
+                                       (trivial-backtrace:print-backtrace
+                                        c :output *error-output*)
+                                       (uiop:quit 3)))
+                                   (skipped-properties (lambda (c)
+                                                         (declare (ignore c))
+                                                         (setq ,failures t))))
+                      ,@forms)))
+       (uiop:quit (cond ((eql :no-change result) 0)
+                        (,failures               2)
+                        (t                       1))))))
+
+(defmacro return-exit (exit &key on-failure)
+  `(values
+    nil
+    (case ,exit
+      (0                              :no-change)
+      (1                              nil)
+      (2 (signal 'skipped-properties) nil)
+      (t                              ,on-failure))))
+
+(defun posix-login-environment (logname home)
+  "Reset the environment after switching UID, or similar, in a :LISP connection.
+Does not currently establish a PAM session."
+  (let ((rootp (zerop (nix:geteuid)))
+        (maybe-preserve '("TERM")))
+    (when rootp
+      (push "SSH_AUTH_SOCK" maybe-preserve))
+    (let ((preserved (loop for var in maybe-preserve
+                           for val = (getenv var)
+                           when val collect var and collect val)))
+      (unless (zerop (foreign-funcall "clearenv" :int))
+        (failed-change "clearenv(3) failed!"))
+      (loop for (var val) on preserved by #'cddr do (setf (getenv var) val)))
+    (setf (getenv "HOME") (drop-trailing-slash (unix-namestring home))
+          (getenv "USER") logname
+          (getenv "LOGNAME") logname
+          (getenv "SHELL") "/bin/sh"
+          (getenv "PATH")
+          (if rootp
+              "/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin"
+              "/usr/local/bin:/bin:/usr/bin"))
+    (uiop:chdir home)))
 
 
 ;;;; Lisp data files
@@ -395,8 +605,37 @@ Should be called soon after fork(2) in child processes."
     `(let* ((,before (and (file-exists-p ,file) (read-file-string ,file)))
             (,data (and ,before (plusp (length ,before))
                         (safe-read-from-string ,before))))
-       (unwind-protect-in-parent (progn ,@forms)
+       (unwind-protect (progn ,@forms)
          (with-open-file
              (stream ,file :direction :output :if-exists :supersede)
            (with-standard-io-syntax
              (prin1 ,data stream)))))))
+
+
+;;;; Streams
+
+(defun stream->input-stream (stream)
+  (etypecase stream
+    (synonym-stream (stream->input-stream
+                     (symbol-value (synonym-stream-symbol stream))))
+    (two-way-stream (two-way-stream-input-stream stream))
+    (stream (and (input-stream-p stream) stream))))
+
+(defun mapc-open-input-streams (function &rest streams)
+  (dolist (stream streams streams)
+    (when-let ((input-stream (stream->input-stream stream)))
+      (when (open-stream-p input-stream)
+        (funcall function input-stream)))))
+
+(defun stream->output-stream (stream)
+  (etypecase stream
+    (synonym-stream (stream->output-stream
+                     (symbol-value (synonym-stream-symbol stream))))
+    (two-way-stream (two-way-stream-output-stream stream))
+    (stream (and (output-stream-p stream) stream))))
+
+(defun mapc-open-output-streams (function &rest streams)
+  (dolist (stream streams streams)
+    (when-let ((output-stream (stream->output-stream stream)))
+      (when (open-stream-p output-stream)
+        (funcall function output-stream)))))
